.TH NAME SECTION 
.\" NAME should be all caps, SECTION should be 1-8, maybe w/ subsection
.\" other parms are allowed: see man(7), man(1)
.SH NAME
cadaver \- programs to do something
.SH SYNOPSIS
.B cadaver
.I "[options] files ..."
.br
.B bar
.I "[options] files ..."
.SH "DESCRIPTION"
This manual page documents briefly the
.BR cadaver ,
and
.B bar
commands.
This manual page was written for the Debian GNU/Linux distribution
because the original program does not have a manual page.
Instead, it has documentation in the GNU Info format; see below.
.PP
.B cadaver
is a program that...
.SH OPTIONS
The programs follow the usual GNU command line syntax, with long
options starting with two dashes (`-').
A summary of options are included below.
For a complete description, see the Info files.
.TP
.B \-h, \-\-help
Show summary of options.
.TP
.B \-v, \-\-version
Show version of program.
.SH "SEE ALSO"
The programs are documented fully by
.IR "The Rise and Fall of a Fooish Bar" ,
available via the Info system.
.SH AUTHOR
This manual page was written by Bernd Eckenfels <ecki@lina.inka.de>,
for the Debian GNU/Linux system (but may be used by others).
